from __future__ import annotations

from argrelay.plugin_interp.FuncArgsInterp import FuncArgsInterp


class DemoInterp(FuncArgsInterp):
    """
    `DemoInterp` is a configured use case of :class:`FuncArgsInterp` to demo auto-completion.

    See FS_55_57_45_04.
    """

    pass
